/*  This file was autogenerated from a template  DO NOT EDIT!!!!
       Changes should be made to the original source (.src) file
*/

/* The purpose of this module is to add faster sort functions
   that are type-specific.  This is done by altering the
   function table for the builtin descriptors. 

   These sorting functions are copied almost directly from numarray 
   with a few modifications (complex comparisons compare the imaginary 
   part if the real parts are equal, for example), and the names
   are changed. 

   The original sorting code is due to Charles R. Harris who wrote
   it for numarray.
*/

/* Quick sort is usually the fastest, but the worst case scenario can
   be slower than the merge and heap sorts.  The merge sort requires
   extra memory and so for large arrays may not be useful. 
 
   The merge sort is *stable*, meaning that equal components
   are unmoved from their entry versions, so it can be used to
   implement lexigraphic sorting on multiple keys. 

   The heap sort is included for completeness. 
*/


#include "Python.h"
#include "numpy/arrayobject.h"

#define PYA_QS_STACK 100
#define SMALL_QUICKSORT 15
#define SMALL_MERGESORT 20
#define STDC_LT(a,b) ((a) < (b))
#define STDC_LE(a,b) ((a) <= (b))
#define STDC_EQ(a,b) ((a) == (b))
#define SWAP(a,b) {SWAP_temp = (b); (b)=(a); (a) = SWAP_temp;}
#define NUMC_LT(p,q) ((((p).real==(q).real) ? ((p).imag < (q).imag): ((p).real < (q).real)))
#define NUMC_LE(p,q) ((((p).real==(q).real) ? ((p).imag <= (q).imag): ((p).real <= (q).real)))
#define NUMC_EQ(p,q) (((p).real==(q).real) && ((p).imag == (q).imag))

static int
BOOL_quicksort(Bool *start, intp num, void *unused)
{
	Bool *pl = start;
	Bool *pr = start + num - 1;
	Bool vp, SWAP_temp;
	Bool *stack[PYA_QS_STACK], **sptr = stack, *pm, *pi, *pj, *pt;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        if (STDC_LT(*pr,*pm)) SWAP(*pr,*pm);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        vp = *pm;
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(*pi,vp));
                                do --pj; while (STDC_LT(vp,*pj));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, *pt);) {
                                *pj-- = *pt--;
                        }
                        *pj = vp;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}

static int
BOOL_aquicksort(Bool *v, intp* tosort, intp num, void *unused)
{
	Bool vp;
	intp *pl, *pr, SWAP_temp;
	intp *stack[PYA_QS_STACK], **sptr=stack, *pm, *pi, *pj, *pt, vi;

	pl = tosort;
	pr = tosort + num - 1;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        if (STDC_LT(v[*pr],v[*pm])) SWAP(*pr,*pm);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        vp = v[*pm];
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(v[*pi],vp));
                                do --pj; while (STDC_LT(vp,v[*pj]));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, v[*pt]);)
				{
					*pj-- = *pt--;
				}
                        *pj = vi;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}


static int
BOOL_heapsort(Bool *start, intp n, void *unused)
{
	
        Bool tmp, *a;
        intp i,j,l;

        /* The array needs to be offset by one for heapsort indexing */
	a = start - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j += 1;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j++;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }
	return 0;
}

static int
BOOL_aheapsort(Bool *v, intp *tosort, intp n, void *unused)
{
	intp *a, i,j,l, tmp;
	/* The arrays need to be offset by one for heapsort indexing */
	a = tosort - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j += 1;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j++;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

	return 0;
}

static void
BOOL_mergesort0(Bool *pl, Bool *pr, Bool *pw)
{
	Bool vp, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                BOOL_mergesort0(pl,pm-1,pw);
                BOOL_mergesort0(pm,pr,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(*pk,*pj)) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pk = pi - 1;\
			    pj > pl && STDC_LT(vp, *pk); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vp;
                }
        }
}

static int
BOOL_mergesort(Bool *start, intp num, void *unused)
{
	Bool *pl, *pr, *pw;

	pl = start; pr = pl + num - 1;
	pw = (Bool *) PyDataMem_NEW(((1+num/2))*sizeof(Bool));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	BOOL_mergesort0(pl, pr, pw);
	PyDataMem_FREE(pw);
	return 0;
}

static void 
BOOL_amergesort0(intp *pl, intp *pr, Bool *v, intp *pw)
{
        Bool vp;
        intp vi, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                BOOL_amergesort0(pl,pm-1,v,pw);
                BOOL_amergesort0(pm,pr,v,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(v[*pk],v[*pj])) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pk = pi - 1;			\
			    pj > pl && STDC_LT(vp, v[*pk]); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vi;
                }
        }
}

static int
BOOL_amergesort(Bool *v, intp *tosort, intp num, void *unused)
{
	intp *pl, *pr, *pw;

	pl = tosort; pr = pl + num - 1;
	pw = PyDimMem_NEW((1+num/2));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	BOOL_amergesort0(pl, pr, v, pw);
	PyDimMem_FREE(pw);
	return 0;
}


static int
BYTE_quicksort(byte *start, intp num, void *unused)
{
	byte *pl = start;
	byte *pr = start + num - 1;
	byte vp, SWAP_temp;
	byte *stack[PYA_QS_STACK], **sptr = stack, *pm, *pi, *pj, *pt;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        if (STDC_LT(*pr,*pm)) SWAP(*pr,*pm);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        vp = *pm;
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(*pi,vp));
                                do --pj; while (STDC_LT(vp,*pj));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, *pt);) {
                                *pj-- = *pt--;
                        }
                        *pj = vp;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}

static int
BYTE_aquicksort(byte *v, intp* tosort, intp num, void *unused)
{
	byte vp;
	intp *pl, *pr, SWAP_temp;
	intp *stack[PYA_QS_STACK], **sptr=stack, *pm, *pi, *pj, *pt, vi;

	pl = tosort;
	pr = tosort + num - 1;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        if (STDC_LT(v[*pr],v[*pm])) SWAP(*pr,*pm);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        vp = v[*pm];
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(v[*pi],vp));
                                do --pj; while (STDC_LT(vp,v[*pj]));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, v[*pt]);)
				{
					*pj-- = *pt--;
				}
                        *pj = vi;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}


static int
BYTE_heapsort(byte *start, intp n, void *unused)
{
	
        byte tmp, *a;
        intp i,j,l;

        /* The array needs to be offset by one for heapsort indexing */
	a = start - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j += 1;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j++;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }
	return 0;
}

static int
BYTE_aheapsort(byte *v, intp *tosort, intp n, void *unused)
{
	intp *a, i,j,l, tmp;
	/* The arrays need to be offset by one for heapsort indexing */
	a = tosort - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j += 1;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j++;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

	return 0;
}

static void
BYTE_mergesort0(byte *pl, byte *pr, byte *pw)
{
	byte vp, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                BYTE_mergesort0(pl,pm-1,pw);
                BYTE_mergesort0(pm,pr,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(*pk,*pj)) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pk = pi - 1;\
			    pj > pl && STDC_LT(vp, *pk); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vp;
                }
        }
}

static int
BYTE_mergesort(byte *start, intp num, void *unused)
{
	byte *pl, *pr, *pw;

	pl = start; pr = pl + num - 1;
	pw = (byte *) PyDataMem_NEW(((1+num/2))*sizeof(byte));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	BYTE_mergesort0(pl, pr, pw);
	PyDataMem_FREE(pw);
	return 0;
}

static void 
BYTE_amergesort0(intp *pl, intp *pr, byte *v, intp *pw)
{
        byte vp;
        intp vi, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                BYTE_amergesort0(pl,pm-1,v,pw);
                BYTE_amergesort0(pm,pr,v,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(v[*pk],v[*pj])) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pk = pi - 1;			\
			    pj > pl && STDC_LT(vp, v[*pk]); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vi;
                }
        }
}

static int
BYTE_amergesort(byte *v, intp *tosort, intp num, void *unused)
{
	intp *pl, *pr, *pw;

	pl = tosort; pr = pl + num - 1;
	pw = PyDimMem_NEW((1+num/2));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	BYTE_amergesort0(pl, pr, v, pw);
	PyDimMem_FREE(pw);
	return 0;
}


static int
UBYTE_quicksort(ubyte *start, intp num, void *unused)
{
	ubyte *pl = start;
	ubyte *pr = start + num - 1;
	ubyte vp, SWAP_temp;
	ubyte *stack[PYA_QS_STACK], **sptr = stack, *pm, *pi, *pj, *pt;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        if (STDC_LT(*pr,*pm)) SWAP(*pr,*pm);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        vp = *pm;
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(*pi,vp));
                                do --pj; while (STDC_LT(vp,*pj));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, *pt);) {
                                *pj-- = *pt--;
                        }
                        *pj = vp;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}

static int
UBYTE_aquicksort(ubyte *v, intp* tosort, intp num, void *unused)
{
	ubyte vp;
	intp *pl, *pr, SWAP_temp;
	intp *stack[PYA_QS_STACK], **sptr=stack, *pm, *pi, *pj, *pt, vi;

	pl = tosort;
	pr = tosort + num - 1;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        if (STDC_LT(v[*pr],v[*pm])) SWAP(*pr,*pm);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        vp = v[*pm];
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(v[*pi],vp));
                                do --pj; while (STDC_LT(vp,v[*pj]));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, v[*pt]);)
				{
					*pj-- = *pt--;
				}
                        *pj = vi;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}


static int
UBYTE_heapsort(ubyte *start, intp n, void *unused)
{
	
        ubyte tmp, *a;
        intp i,j,l;

        /* The array needs to be offset by one for heapsort indexing */
	a = start - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j += 1;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j++;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }
	return 0;
}

static int
UBYTE_aheapsort(ubyte *v, intp *tosort, intp n, void *unused)
{
	intp *a, i,j,l, tmp;
	/* The arrays need to be offset by one for heapsort indexing */
	a = tosort - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j += 1;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j++;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

	return 0;
}

static void
UBYTE_mergesort0(ubyte *pl, ubyte *pr, ubyte *pw)
{
	ubyte vp, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                UBYTE_mergesort0(pl,pm-1,pw);
                UBYTE_mergesort0(pm,pr,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(*pk,*pj)) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pk = pi - 1;\
			    pj > pl && STDC_LT(vp, *pk); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vp;
                }
        }
}

static int
UBYTE_mergesort(ubyte *start, intp num, void *unused)
{
	ubyte *pl, *pr, *pw;

	pl = start; pr = pl + num - 1;
	pw = (ubyte *) PyDataMem_NEW(((1+num/2))*sizeof(ubyte));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	UBYTE_mergesort0(pl, pr, pw);
	PyDataMem_FREE(pw);
	return 0;
}

static void 
UBYTE_amergesort0(intp *pl, intp *pr, ubyte *v, intp *pw)
{
        ubyte vp;
        intp vi, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                UBYTE_amergesort0(pl,pm-1,v,pw);
                UBYTE_amergesort0(pm,pr,v,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(v[*pk],v[*pj])) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pk = pi - 1;			\
			    pj > pl && STDC_LT(vp, v[*pk]); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vi;
                }
        }
}

static int
UBYTE_amergesort(ubyte *v, intp *tosort, intp num, void *unused)
{
	intp *pl, *pr, *pw;

	pl = tosort; pr = pl + num - 1;
	pw = PyDimMem_NEW((1+num/2));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	UBYTE_amergesort0(pl, pr, v, pw);
	PyDimMem_FREE(pw);
	return 0;
}


static int
SHORT_quicksort(short *start, intp num, void *unused)
{
	short *pl = start;
	short *pr = start + num - 1;
	short vp, SWAP_temp;
	short *stack[PYA_QS_STACK], **sptr = stack, *pm, *pi, *pj, *pt;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        if (STDC_LT(*pr,*pm)) SWAP(*pr,*pm);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        vp = *pm;
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(*pi,vp));
                                do --pj; while (STDC_LT(vp,*pj));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, *pt);) {
                                *pj-- = *pt--;
                        }
                        *pj = vp;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}

static int
SHORT_aquicksort(short *v, intp* tosort, intp num, void *unused)
{
	short vp;
	intp *pl, *pr, SWAP_temp;
	intp *stack[PYA_QS_STACK], **sptr=stack, *pm, *pi, *pj, *pt, vi;

	pl = tosort;
	pr = tosort + num - 1;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        if (STDC_LT(v[*pr],v[*pm])) SWAP(*pr,*pm);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        vp = v[*pm];
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(v[*pi],vp));
                                do --pj; while (STDC_LT(vp,v[*pj]));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, v[*pt]);)
				{
					*pj-- = *pt--;
				}
                        *pj = vi;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}


static int
SHORT_heapsort(short *start, intp n, void *unused)
{
	
        short tmp, *a;
        intp i,j,l;

        /* The array needs to be offset by one for heapsort indexing */
	a = start - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j += 1;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j++;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }
	return 0;
}

static int
SHORT_aheapsort(short *v, intp *tosort, intp n, void *unused)
{
	intp *a, i,j,l, tmp;
	/* The arrays need to be offset by one for heapsort indexing */
	a = tosort - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j += 1;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j++;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

	return 0;
}

static void
SHORT_mergesort0(short *pl, short *pr, short *pw)
{
	short vp, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                SHORT_mergesort0(pl,pm-1,pw);
                SHORT_mergesort0(pm,pr,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(*pk,*pj)) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pk = pi - 1;\
			    pj > pl && STDC_LT(vp, *pk); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vp;
                }
        }
}

static int
SHORT_mergesort(short *start, intp num, void *unused)
{
	short *pl, *pr, *pw;

	pl = start; pr = pl + num - 1;
	pw = (short *) PyDataMem_NEW(((1+num/2))*sizeof(short));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	SHORT_mergesort0(pl, pr, pw);
	PyDataMem_FREE(pw);
	return 0;
}

static void 
SHORT_amergesort0(intp *pl, intp *pr, short *v, intp *pw)
{
        short vp;
        intp vi, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                SHORT_amergesort0(pl,pm-1,v,pw);
                SHORT_amergesort0(pm,pr,v,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(v[*pk],v[*pj])) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pk = pi - 1;			\
			    pj > pl && STDC_LT(vp, v[*pk]); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vi;
                }
        }
}

static int
SHORT_amergesort(short *v, intp *tosort, intp num, void *unused)
{
	intp *pl, *pr, *pw;

	pl = tosort; pr = pl + num - 1;
	pw = PyDimMem_NEW((1+num/2));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	SHORT_amergesort0(pl, pr, v, pw);
	PyDimMem_FREE(pw);
	return 0;
}


static int
USHORT_quicksort(ushort *start, intp num, void *unused)
{
	ushort *pl = start;
	ushort *pr = start + num - 1;
	ushort vp, SWAP_temp;
	ushort *stack[PYA_QS_STACK], **sptr = stack, *pm, *pi, *pj, *pt;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        if (STDC_LT(*pr,*pm)) SWAP(*pr,*pm);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        vp = *pm;
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(*pi,vp));
                                do --pj; while (STDC_LT(vp,*pj));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, *pt);) {
                                *pj-- = *pt--;
                        }
                        *pj = vp;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}

static int
USHORT_aquicksort(ushort *v, intp* tosort, intp num, void *unused)
{
	ushort vp;
	intp *pl, *pr, SWAP_temp;
	intp *stack[PYA_QS_STACK], **sptr=stack, *pm, *pi, *pj, *pt, vi;

	pl = tosort;
	pr = tosort + num - 1;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        if (STDC_LT(v[*pr],v[*pm])) SWAP(*pr,*pm);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        vp = v[*pm];
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(v[*pi],vp));
                                do --pj; while (STDC_LT(vp,v[*pj]));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, v[*pt]);)
				{
					*pj-- = *pt--;
				}
                        *pj = vi;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}


static int
USHORT_heapsort(ushort *start, intp n, void *unused)
{
	
        ushort tmp, *a;
        intp i,j,l;

        /* The array needs to be offset by one for heapsort indexing */
	a = start - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j += 1;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j++;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }
	return 0;
}

static int
USHORT_aheapsort(ushort *v, intp *tosort, intp n, void *unused)
{
	intp *a, i,j,l, tmp;
	/* The arrays need to be offset by one for heapsort indexing */
	a = tosort - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j += 1;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j++;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

	return 0;
}

static void
USHORT_mergesort0(ushort *pl, ushort *pr, ushort *pw)
{
	ushort vp, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                USHORT_mergesort0(pl,pm-1,pw);
                USHORT_mergesort0(pm,pr,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(*pk,*pj)) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pk = pi - 1;\
			    pj > pl && STDC_LT(vp, *pk); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vp;
                }
        }
}

static int
USHORT_mergesort(ushort *start, intp num, void *unused)
{
	ushort *pl, *pr, *pw;

	pl = start; pr = pl + num - 1;
	pw = (ushort *) PyDataMem_NEW(((1+num/2))*sizeof(ushort));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	USHORT_mergesort0(pl, pr, pw);
	PyDataMem_FREE(pw);
	return 0;
}

static void 
USHORT_amergesort0(intp *pl, intp *pr, ushort *v, intp *pw)
{
        ushort vp;
        intp vi, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                USHORT_amergesort0(pl,pm-1,v,pw);
                USHORT_amergesort0(pm,pr,v,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(v[*pk],v[*pj])) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pk = pi - 1;			\
			    pj > pl && STDC_LT(vp, v[*pk]); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vi;
                }
        }
}

static int
USHORT_amergesort(ushort *v, intp *tosort, intp num, void *unused)
{
	intp *pl, *pr, *pw;

	pl = tosort; pr = pl + num - 1;
	pw = PyDimMem_NEW((1+num/2));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	USHORT_amergesort0(pl, pr, v, pw);
	PyDimMem_FREE(pw);
	return 0;
}


static int
INT_quicksort(int *start, intp num, void *unused)
{
	int *pl = start;
	int *pr = start + num - 1;
	int vp, SWAP_temp;
	int *stack[PYA_QS_STACK], **sptr = stack, *pm, *pi, *pj, *pt;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        if (STDC_LT(*pr,*pm)) SWAP(*pr,*pm);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        vp = *pm;
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(*pi,vp));
                                do --pj; while (STDC_LT(vp,*pj));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, *pt);) {
                                *pj-- = *pt--;
                        }
                        *pj = vp;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}

static int
INT_aquicksort(int *v, intp* tosort, intp num, void *unused)
{
	int vp;
	intp *pl, *pr, SWAP_temp;
	intp *stack[PYA_QS_STACK], **sptr=stack, *pm, *pi, *pj, *pt, vi;

	pl = tosort;
	pr = tosort + num - 1;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        if (STDC_LT(v[*pr],v[*pm])) SWAP(*pr,*pm);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        vp = v[*pm];
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(v[*pi],vp));
                                do --pj; while (STDC_LT(vp,v[*pj]));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, v[*pt]);)
				{
					*pj-- = *pt--;
				}
                        *pj = vi;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}


static int
INT_heapsort(int *start, intp n, void *unused)
{
	
        int tmp, *a;
        intp i,j,l;

        /* The array needs to be offset by one for heapsort indexing */
	a = start - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j += 1;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j++;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }
	return 0;
}

static int
INT_aheapsort(int *v, intp *tosort, intp n, void *unused)
{
	intp *a, i,j,l, tmp;
	/* The arrays need to be offset by one for heapsort indexing */
	a = tosort - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j += 1;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j++;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

	return 0;
}

static void
INT_mergesort0(int *pl, int *pr, int *pw)
{
	int vp, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                INT_mergesort0(pl,pm-1,pw);
                INT_mergesort0(pm,pr,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(*pk,*pj)) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pk = pi - 1;\
			    pj > pl && STDC_LT(vp, *pk); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vp;
                }
        }
}

static int
INT_mergesort(int *start, intp num, void *unused)
{
	int *pl, *pr, *pw;

	pl = start; pr = pl + num - 1;
	pw = (int *) PyDataMem_NEW(((1+num/2))*sizeof(int));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	INT_mergesort0(pl, pr, pw);
	PyDataMem_FREE(pw);
	return 0;
}

static void 
INT_amergesort0(intp *pl, intp *pr, int *v, intp *pw)
{
        int vp;
        intp vi, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                INT_amergesort0(pl,pm-1,v,pw);
                INT_amergesort0(pm,pr,v,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(v[*pk],v[*pj])) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pk = pi - 1;			\
			    pj > pl && STDC_LT(vp, v[*pk]); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vi;
                }
        }
}

static int
INT_amergesort(int *v, intp *tosort, intp num, void *unused)
{
	intp *pl, *pr, *pw;

	pl = tosort; pr = pl + num - 1;
	pw = PyDimMem_NEW((1+num/2));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	INT_amergesort0(pl, pr, v, pw);
	PyDimMem_FREE(pw);
	return 0;
}


static int
UINT_quicksort(uint *start, intp num, void *unused)
{
	uint *pl = start;
	uint *pr = start + num - 1;
	uint vp, SWAP_temp;
	uint *stack[PYA_QS_STACK], **sptr = stack, *pm, *pi, *pj, *pt;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        if (STDC_LT(*pr,*pm)) SWAP(*pr,*pm);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        vp = *pm;
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(*pi,vp));
                                do --pj; while (STDC_LT(vp,*pj));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, *pt);) {
                                *pj-- = *pt--;
                        }
                        *pj = vp;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}

static int
UINT_aquicksort(uint *v, intp* tosort, intp num, void *unused)
{
	uint vp;
	intp *pl, *pr, SWAP_temp;
	intp *stack[PYA_QS_STACK], **sptr=stack, *pm, *pi, *pj, *pt, vi;

	pl = tosort;
	pr = tosort + num - 1;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        if (STDC_LT(v[*pr],v[*pm])) SWAP(*pr,*pm);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        vp = v[*pm];
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(v[*pi],vp));
                                do --pj; while (STDC_LT(vp,v[*pj]));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, v[*pt]);)
				{
					*pj-- = *pt--;
				}
                        *pj = vi;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}


static int
UINT_heapsort(uint *start, intp n, void *unused)
{
	
        uint tmp, *a;
        intp i,j,l;

        /* The array needs to be offset by one for heapsort indexing */
	a = start - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j += 1;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j++;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }
	return 0;
}

static int
UINT_aheapsort(uint *v, intp *tosort, intp n, void *unused)
{
	intp *a, i,j,l, tmp;
	/* The arrays need to be offset by one for heapsort indexing */
	a = tosort - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j += 1;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j++;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

	return 0;
}

static void
UINT_mergesort0(uint *pl, uint *pr, uint *pw)
{
	uint vp, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                UINT_mergesort0(pl,pm-1,pw);
                UINT_mergesort0(pm,pr,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(*pk,*pj)) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pk = pi - 1;\
			    pj > pl && STDC_LT(vp, *pk); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vp;
                }
        }
}

static int
UINT_mergesort(uint *start, intp num, void *unused)
{
	uint *pl, *pr, *pw;

	pl = start; pr = pl + num - 1;
	pw = (uint *) PyDataMem_NEW(((1+num/2))*sizeof(uint));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	UINT_mergesort0(pl, pr, pw);
	PyDataMem_FREE(pw);
	return 0;
}

static void 
UINT_amergesort0(intp *pl, intp *pr, uint *v, intp *pw)
{
        uint vp;
        intp vi, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                UINT_amergesort0(pl,pm-1,v,pw);
                UINT_amergesort0(pm,pr,v,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(v[*pk],v[*pj])) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pk = pi - 1;			\
			    pj > pl && STDC_LT(vp, v[*pk]); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vi;
                }
        }
}

static int
UINT_amergesort(uint *v, intp *tosort, intp num, void *unused)
{
	intp *pl, *pr, *pw;

	pl = tosort; pr = pl + num - 1;
	pw = PyDimMem_NEW((1+num/2));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	UINT_amergesort0(pl, pr, v, pw);
	PyDimMem_FREE(pw);
	return 0;
}


static int
LONG_quicksort(long *start, intp num, void *unused)
{
	long *pl = start;
	long *pr = start + num - 1;
	long vp, SWAP_temp;
	long *stack[PYA_QS_STACK], **sptr = stack, *pm, *pi, *pj, *pt;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        if (STDC_LT(*pr,*pm)) SWAP(*pr,*pm);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        vp = *pm;
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(*pi,vp));
                                do --pj; while (STDC_LT(vp,*pj));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, *pt);) {
                                *pj-- = *pt--;
                        }
                        *pj = vp;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}

static int
LONG_aquicksort(long *v, intp* tosort, intp num, void *unused)
{
	long vp;
	intp *pl, *pr, SWAP_temp;
	intp *stack[PYA_QS_STACK], **sptr=stack, *pm, *pi, *pj, *pt, vi;

	pl = tosort;
	pr = tosort + num - 1;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        if (STDC_LT(v[*pr],v[*pm])) SWAP(*pr,*pm);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        vp = v[*pm];
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(v[*pi],vp));
                                do --pj; while (STDC_LT(vp,v[*pj]));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, v[*pt]);)
				{
					*pj-- = *pt--;
				}
                        *pj = vi;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}


static int
LONG_heapsort(long *start, intp n, void *unused)
{
	
        long tmp, *a;
        intp i,j,l;

        /* The array needs to be offset by one for heapsort indexing */
	a = start - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j += 1;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j++;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }
	return 0;
}

static int
LONG_aheapsort(long *v, intp *tosort, intp n, void *unused)
{
	intp *a, i,j,l, tmp;
	/* The arrays need to be offset by one for heapsort indexing */
	a = tosort - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j += 1;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j++;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

	return 0;
}

static void
LONG_mergesort0(long *pl, long *pr, long *pw)
{
	long vp, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                LONG_mergesort0(pl,pm-1,pw);
                LONG_mergesort0(pm,pr,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(*pk,*pj)) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pk = pi - 1;\
			    pj > pl && STDC_LT(vp, *pk); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vp;
                }
        }
}

static int
LONG_mergesort(long *start, intp num, void *unused)
{
	long *pl, *pr, *pw;

	pl = start; pr = pl + num - 1;
	pw = (long *) PyDataMem_NEW(((1+num/2))*sizeof(long));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	LONG_mergesort0(pl, pr, pw);
	PyDataMem_FREE(pw);
	return 0;
}

static void 
LONG_amergesort0(intp *pl, intp *pr, long *v, intp *pw)
{
        long vp;
        intp vi, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                LONG_amergesort0(pl,pm-1,v,pw);
                LONG_amergesort0(pm,pr,v,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(v[*pk],v[*pj])) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pk = pi - 1;			\
			    pj > pl && STDC_LT(vp, v[*pk]); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vi;
                }
        }
}

static int
LONG_amergesort(long *v, intp *tosort, intp num, void *unused)
{
	intp *pl, *pr, *pw;

	pl = tosort; pr = pl + num - 1;
	pw = PyDimMem_NEW((1+num/2));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	LONG_amergesort0(pl, pr, v, pw);
	PyDimMem_FREE(pw);
	return 0;
}


static int
ULONG_quicksort(ulong *start, intp num, void *unused)
{
	ulong *pl = start;
	ulong *pr = start + num - 1;
	ulong vp, SWAP_temp;
	ulong *stack[PYA_QS_STACK], **sptr = stack, *pm, *pi, *pj, *pt;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        if (STDC_LT(*pr,*pm)) SWAP(*pr,*pm);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        vp = *pm;
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(*pi,vp));
                                do --pj; while (STDC_LT(vp,*pj));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, *pt);) {
                                *pj-- = *pt--;
                        }
                        *pj = vp;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}

static int
ULONG_aquicksort(ulong *v, intp* tosort, intp num, void *unused)
{
	ulong vp;
	intp *pl, *pr, SWAP_temp;
	intp *stack[PYA_QS_STACK], **sptr=stack, *pm, *pi, *pj, *pt, vi;

	pl = tosort;
	pr = tosort + num - 1;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        if (STDC_LT(v[*pr],v[*pm])) SWAP(*pr,*pm);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        vp = v[*pm];
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(v[*pi],vp));
                                do --pj; while (STDC_LT(vp,v[*pj]));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, v[*pt]);)
				{
					*pj-- = *pt--;
				}
                        *pj = vi;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}


static int
ULONG_heapsort(ulong *start, intp n, void *unused)
{
	
        ulong tmp, *a;
        intp i,j,l;

        /* The array needs to be offset by one for heapsort indexing */
	a = start - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j += 1;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j++;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }
	return 0;
}

static int
ULONG_aheapsort(ulong *v, intp *tosort, intp n, void *unused)
{
	intp *a, i,j,l, tmp;
	/* The arrays need to be offset by one for heapsort indexing */
	a = tosort - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j += 1;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j++;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

	return 0;
}

static void
ULONG_mergesort0(ulong *pl, ulong *pr, ulong *pw)
{
	ulong vp, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                ULONG_mergesort0(pl,pm-1,pw);
                ULONG_mergesort0(pm,pr,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(*pk,*pj)) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pk = pi - 1;\
			    pj > pl && STDC_LT(vp, *pk); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vp;
                }
        }
}

static int
ULONG_mergesort(ulong *start, intp num, void *unused)
{
	ulong *pl, *pr, *pw;

	pl = start; pr = pl + num - 1;
	pw = (ulong *) PyDataMem_NEW(((1+num/2))*sizeof(ulong));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	ULONG_mergesort0(pl, pr, pw);
	PyDataMem_FREE(pw);
	return 0;
}

static void 
ULONG_amergesort0(intp *pl, intp *pr, ulong *v, intp *pw)
{
        ulong vp;
        intp vi, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                ULONG_amergesort0(pl,pm-1,v,pw);
                ULONG_amergesort0(pm,pr,v,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(v[*pk],v[*pj])) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pk = pi - 1;			\
			    pj > pl && STDC_LT(vp, v[*pk]); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vi;
                }
        }
}

static int
ULONG_amergesort(ulong *v, intp *tosort, intp num, void *unused)
{
	intp *pl, *pr, *pw;

	pl = tosort; pr = pl + num - 1;
	pw = PyDimMem_NEW((1+num/2));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	ULONG_amergesort0(pl, pr, v, pw);
	PyDimMem_FREE(pw);
	return 0;
}


static int
LONGLONG_quicksort(longlong *start, intp num, void *unused)
{
	longlong *pl = start;
	longlong *pr = start + num - 1;
	longlong vp, SWAP_temp;
	longlong *stack[PYA_QS_STACK], **sptr = stack, *pm, *pi, *pj, *pt;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        if (STDC_LT(*pr,*pm)) SWAP(*pr,*pm);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        vp = *pm;
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(*pi,vp));
                                do --pj; while (STDC_LT(vp,*pj));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, *pt);) {
                                *pj-- = *pt--;
                        }
                        *pj = vp;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}

static int
LONGLONG_aquicksort(longlong *v, intp* tosort, intp num, void *unused)
{
	longlong vp;
	intp *pl, *pr, SWAP_temp;
	intp *stack[PYA_QS_STACK], **sptr=stack, *pm, *pi, *pj, *pt, vi;

	pl = tosort;
	pr = tosort + num - 1;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        if (STDC_LT(v[*pr],v[*pm])) SWAP(*pr,*pm);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        vp = v[*pm];
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(v[*pi],vp));
                                do --pj; while (STDC_LT(vp,v[*pj]));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, v[*pt]);)
				{
					*pj-- = *pt--;
				}
                        *pj = vi;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}


static int
LONGLONG_heapsort(longlong *start, intp n, void *unused)
{
	
        longlong tmp, *a;
        intp i,j,l;

        /* The array needs to be offset by one for heapsort indexing */
	a = start - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j += 1;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j++;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }
	return 0;
}

static int
LONGLONG_aheapsort(longlong *v, intp *tosort, intp n, void *unused)
{
	intp *a, i,j,l, tmp;
	/* The arrays need to be offset by one for heapsort indexing */
	a = tosort - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j += 1;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j++;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

	return 0;
}

static void
LONGLONG_mergesort0(longlong *pl, longlong *pr, longlong *pw)
{
	longlong vp, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                LONGLONG_mergesort0(pl,pm-1,pw);
                LONGLONG_mergesort0(pm,pr,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(*pk,*pj)) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pk = pi - 1;\
			    pj > pl && STDC_LT(vp, *pk); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vp;
                }
        }
}

static int
LONGLONG_mergesort(longlong *start, intp num, void *unused)
{
	longlong *pl, *pr, *pw;

	pl = start; pr = pl + num - 1;
	pw = (longlong *) PyDataMem_NEW(((1+num/2))*sizeof(longlong));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	LONGLONG_mergesort0(pl, pr, pw);
	PyDataMem_FREE(pw);
	return 0;
}

static void 
LONGLONG_amergesort0(intp *pl, intp *pr, longlong *v, intp *pw)
{
        longlong vp;
        intp vi, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                LONGLONG_amergesort0(pl,pm-1,v,pw);
                LONGLONG_amergesort0(pm,pr,v,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(v[*pk],v[*pj])) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pk = pi - 1;			\
			    pj > pl && STDC_LT(vp, v[*pk]); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vi;
                }
        }
}

static int
LONGLONG_amergesort(longlong *v, intp *tosort, intp num, void *unused)
{
	intp *pl, *pr, *pw;

	pl = tosort; pr = pl + num - 1;
	pw = PyDimMem_NEW((1+num/2));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	LONGLONG_amergesort0(pl, pr, v, pw);
	PyDimMem_FREE(pw);
	return 0;
}


static int
ULONGLONG_quicksort(ulonglong *start, intp num, void *unused)
{
	ulonglong *pl = start;
	ulonglong *pr = start + num - 1;
	ulonglong vp, SWAP_temp;
	ulonglong *stack[PYA_QS_STACK], **sptr = stack, *pm, *pi, *pj, *pt;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        if (STDC_LT(*pr,*pm)) SWAP(*pr,*pm);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        vp = *pm;
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(*pi,vp));
                                do --pj; while (STDC_LT(vp,*pj));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, *pt);) {
                                *pj-- = *pt--;
                        }
                        *pj = vp;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}

static int
ULONGLONG_aquicksort(ulonglong *v, intp* tosort, intp num, void *unused)
{
	ulonglong vp;
	intp *pl, *pr, SWAP_temp;
	intp *stack[PYA_QS_STACK], **sptr=stack, *pm, *pi, *pj, *pt, vi;

	pl = tosort;
	pr = tosort + num - 1;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        if (STDC_LT(v[*pr],v[*pm])) SWAP(*pr,*pm);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        vp = v[*pm];
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(v[*pi],vp));
                                do --pj; while (STDC_LT(vp,v[*pj]));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, v[*pt]);)
				{
					*pj-- = *pt--;
				}
                        *pj = vi;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}


static int
ULONGLONG_heapsort(ulonglong *start, intp n, void *unused)
{
	
        ulonglong tmp, *a;
        intp i,j,l;

        /* The array needs to be offset by one for heapsort indexing */
	a = start - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j += 1;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j++;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }
	return 0;
}

static int
ULONGLONG_aheapsort(ulonglong *v, intp *tosort, intp n, void *unused)
{
	intp *a, i,j,l, tmp;
	/* The arrays need to be offset by one for heapsort indexing */
	a = tosort - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j += 1;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j++;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

	return 0;
}

static void
ULONGLONG_mergesort0(ulonglong *pl, ulonglong *pr, ulonglong *pw)
{
	ulonglong vp, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                ULONGLONG_mergesort0(pl,pm-1,pw);
                ULONGLONG_mergesort0(pm,pr,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(*pk,*pj)) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pk = pi - 1;\
			    pj > pl && STDC_LT(vp, *pk); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vp;
                }
        }
}

static int
ULONGLONG_mergesort(ulonglong *start, intp num, void *unused)
{
	ulonglong *pl, *pr, *pw;

	pl = start; pr = pl + num - 1;
	pw = (ulonglong *) PyDataMem_NEW(((1+num/2))*sizeof(ulonglong));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	ULONGLONG_mergesort0(pl, pr, pw);
	PyDataMem_FREE(pw);
	return 0;
}

static void 
ULONGLONG_amergesort0(intp *pl, intp *pr, ulonglong *v, intp *pw)
{
        ulonglong vp;
        intp vi, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                ULONGLONG_amergesort0(pl,pm-1,v,pw);
                ULONGLONG_amergesort0(pm,pr,v,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(v[*pk],v[*pj])) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pk = pi - 1;			\
			    pj > pl && STDC_LT(vp, v[*pk]); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vi;
                }
        }
}

static int
ULONGLONG_amergesort(ulonglong *v, intp *tosort, intp num, void *unused)
{
	intp *pl, *pr, *pw;

	pl = tosort; pr = pl + num - 1;
	pw = PyDimMem_NEW((1+num/2));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	ULONGLONG_amergesort0(pl, pr, v, pw);
	PyDimMem_FREE(pw);
	return 0;
}


static int
FLOAT_quicksort(float *start, intp num, void *unused)
{
	float *pl = start;
	float *pr = start + num - 1;
	float vp, SWAP_temp;
	float *stack[PYA_QS_STACK], **sptr = stack, *pm, *pi, *pj, *pt;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        if (STDC_LT(*pr,*pm)) SWAP(*pr,*pm);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        vp = *pm;
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(*pi,vp));
                                do --pj; while (STDC_LT(vp,*pj));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, *pt);) {
                                *pj-- = *pt--;
                        }
                        *pj = vp;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}

static int
FLOAT_aquicksort(float *v, intp* tosort, intp num, void *unused)
{
	float vp;
	intp *pl, *pr, SWAP_temp;
	intp *stack[PYA_QS_STACK], **sptr=stack, *pm, *pi, *pj, *pt, vi;

	pl = tosort;
	pr = tosort + num - 1;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        if (STDC_LT(v[*pr],v[*pm])) SWAP(*pr,*pm);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        vp = v[*pm];
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(v[*pi],vp));
                                do --pj; while (STDC_LT(vp,v[*pj]));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, v[*pt]);)
				{
					*pj-- = *pt--;
				}
                        *pj = vi;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}


static int
FLOAT_heapsort(float *start, intp n, void *unused)
{
	
        float tmp, *a;
        intp i,j,l;

        /* The array needs to be offset by one for heapsort indexing */
	a = start - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j += 1;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j++;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }
	return 0;
}

static int
FLOAT_aheapsort(float *v, intp *tosort, intp n, void *unused)
{
	intp *a, i,j,l, tmp;
	/* The arrays need to be offset by one for heapsort indexing */
	a = tosort - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j += 1;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j++;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

	return 0;
}

static void
FLOAT_mergesort0(float *pl, float *pr, float *pw)
{
	float vp, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                FLOAT_mergesort0(pl,pm-1,pw);
                FLOAT_mergesort0(pm,pr,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(*pk,*pj)) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pk = pi - 1;\
			    pj > pl && STDC_LT(vp, *pk); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vp;
                }
        }
}

static int
FLOAT_mergesort(float *start, intp num, void *unused)
{
	float *pl, *pr, *pw;

	pl = start; pr = pl + num - 1;
	pw = (float *) PyDataMem_NEW(((1+num/2))*sizeof(float));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	FLOAT_mergesort0(pl, pr, pw);
	PyDataMem_FREE(pw);
	return 0;
}

static void 
FLOAT_amergesort0(intp *pl, intp *pr, float *v, intp *pw)
{
        float vp;
        intp vi, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                FLOAT_amergesort0(pl,pm-1,v,pw);
                FLOAT_amergesort0(pm,pr,v,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(v[*pk],v[*pj])) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pk = pi - 1;			\
			    pj > pl && STDC_LT(vp, v[*pk]); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vi;
                }
        }
}

static int
FLOAT_amergesort(float *v, intp *tosort, intp num, void *unused)
{
	intp *pl, *pr, *pw;

	pl = tosort; pr = pl + num - 1;
	pw = PyDimMem_NEW((1+num/2));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	FLOAT_amergesort0(pl, pr, v, pw);
	PyDimMem_FREE(pw);
	return 0;
}


static int
DOUBLE_quicksort(double *start, intp num, void *unused)
{
	double *pl = start;
	double *pr = start + num - 1;
	double vp, SWAP_temp;
	double *stack[PYA_QS_STACK], **sptr = stack, *pm, *pi, *pj, *pt;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        if (STDC_LT(*pr,*pm)) SWAP(*pr,*pm);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        vp = *pm;
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(*pi,vp));
                                do --pj; while (STDC_LT(vp,*pj));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, *pt);) {
                                *pj-- = *pt--;
                        }
                        *pj = vp;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}

static int
DOUBLE_aquicksort(double *v, intp* tosort, intp num, void *unused)
{
	double vp;
	intp *pl, *pr, SWAP_temp;
	intp *stack[PYA_QS_STACK], **sptr=stack, *pm, *pi, *pj, *pt, vi;

	pl = tosort;
	pr = tosort + num - 1;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        if (STDC_LT(v[*pr],v[*pm])) SWAP(*pr,*pm);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        vp = v[*pm];
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(v[*pi],vp));
                                do --pj; while (STDC_LT(vp,v[*pj]));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, v[*pt]);)
				{
					*pj-- = *pt--;
				}
                        *pj = vi;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}


static int
DOUBLE_heapsort(double *start, intp n, void *unused)
{
	
        double tmp, *a;
        intp i,j,l;

        /* The array needs to be offset by one for heapsort indexing */
	a = start - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j += 1;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j++;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }
	return 0;
}

static int
DOUBLE_aheapsort(double *v, intp *tosort, intp n, void *unused)
{
	intp *a, i,j,l, tmp;
	/* The arrays need to be offset by one for heapsort indexing */
	a = tosort - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j += 1;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j++;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

	return 0;
}

static void
DOUBLE_mergesort0(double *pl, double *pr, double *pw)
{
	double vp, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                DOUBLE_mergesort0(pl,pm-1,pw);
                DOUBLE_mergesort0(pm,pr,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(*pk,*pj)) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pk = pi - 1;\
			    pj > pl && STDC_LT(vp, *pk); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vp;
                }
        }
}

static int
DOUBLE_mergesort(double *start, intp num, void *unused)
{
	double *pl, *pr, *pw;

	pl = start; pr = pl + num - 1;
	pw = (double *) PyDataMem_NEW(((1+num/2))*sizeof(double));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	DOUBLE_mergesort0(pl, pr, pw);
	PyDataMem_FREE(pw);
	return 0;
}

static void 
DOUBLE_amergesort0(intp *pl, intp *pr, double *v, intp *pw)
{
        double vp;
        intp vi, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                DOUBLE_amergesort0(pl,pm-1,v,pw);
                DOUBLE_amergesort0(pm,pr,v,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(v[*pk],v[*pj])) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pk = pi - 1;			\
			    pj > pl && STDC_LT(vp, v[*pk]); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vi;
                }
        }
}

static int
DOUBLE_amergesort(double *v, intp *tosort, intp num, void *unused)
{
	intp *pl, *pr, *pw;

	pl = tosort; pr = pl + num - 1;
	pw = PyDimMem_NEW((1+num/2));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	DOUBLE_amergesort0(pl, pr, v, pw);
	PyDimMem_FREE(pw);
	return 0;
}


static int
LONGDOUBLE_quicksort(longdouble *start, intp num, void *unused)
{
	longdouble *pl = start;
	longdouble *pr = start + num - 1;
	longdouble vp, SWAP_temp;
	longdouble *stack[PYA_QS_STACK], **sptr = stack, *pm, *pi, *pj, *pt;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        if (STDC_LT(*pr,*pm)) SWAP(*pr,*pm);
                        if (STDC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        vp = *pm;
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(*pi,vp));
                                do --pj; while (STDC_LT(vp,*pj));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, *pt);) {
                                *pj-- = *pt--;
                        }
                        *pj = vp;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}

static int
LONGDOUBLE_aquicksort(longdouble *v, intp* tosort, intp num, void *unused)
{
	longdouble vp;
	intp *pl, *pr, SWAP_temp;
	intp *stack[PYA_QS_STACK], **sptr=stack, *pm, *pi, *pj, *pt, vi;

	pl = tosort;
	pr = tosort + num - 1;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        if (STDC_LT(v[*pr],v[*pm])) SWAP(*pr,*pm);
                        if (STDC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        vp = v[*pm];
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (STDC_LT(v[*pi],vp));
                                do --pj; while (STDC_LT(vp,v[*pj]));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && STDC_LT(vp, v[*pt]);)
				{
					*pj-- = *pt--;
				}
                        *pj = vi;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}


static int
LONGDOUBLE_heapsort(longdouble *start, intp n, void *unused)
{
	
        longdouble tmp, *a;
        intp i,j,l;

        /* The array needs to be offset by one for heapsort indexing */
	a = start - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j += 1;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(a[j], a[j+1]))
                                j++;
                        if (STDC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }
	return 0;
}

static int
LONGDOUBLE_aheapsort(longdouble *v, intp *tosort, intp n, void *unused)
{
	intp *a, i,j,l, tmp;
	/* The arrays need to be offset by one for heapsort indexing */
	a = tosort - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j += 1;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && STDC_LT(v[a[j]], v[a[j+1]]))
                                j++;
                        if (STDC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

	return 0;
}

static void
LONGDOUBLE_mergesort0(longdouble *pl, longdouble *pr, longdouble *pw)
{
	longdouble vp, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                LONGDOUBLE_mergesort0(pl,pm-1,pw);
                LONGDOUBLE_mergesort0(pm,pr,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(*pk,*pj)) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pk = pi - 1;\
			    pj > pl && STDC_LT(vp, *pk); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vp;
                }
        }
}

static int
LONGDOUBLE_mergesort(longdouble *start, intp num, void *unused)
{
	longdouble *pl, *pr, *pw;

	pl = start; pr = pl + num - 1;
	pw = (longdouble *) PyDataMem_NEW(((1+num/2))*sizeof(longdouble));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	LONGDOUBLE_mergesort0(pl, pr, pw);
	PyDataMem_FREE(pw);
	return 0;
}

static void 
LONGDOUBLE_amergesort0(intp *pl, intp *pr, longdouble *v, intp *pw)
{
        longdouble vp;
        intp vi, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                LONGDOUBLE_amergesort0(pl,pm-1,v,pw);
                LONGDOUBLE_amergesort0(pm,pr,v,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (STDC_LE(v[*pk],v[*pj])) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pk = pi - 1;			\
			    pj > pl && STDC_LT(vp, v[*pk]); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vi;
                }
        }
}

static int
LONGDOUBLE_amergesort(longdouble *v, intp *tosort, intp num, void *unused)
{
	intp *pl, *pr, *pw;

	pl = tosort; pr = pl + num - 1;
	pw = PyDimMem_NEW((1+num/2));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	LONGDOUBLE_amergesort0(pl, pr, v, pw);
	PyDimMem_FREE(pw);
	return 0;
}


static int
CFLOAT_quicksort(cfloat *start, intp num, void *unused)
{
	cfloat *pl = start;
	cfloat *pr = start + num - 1;
	cfloat vp, SWAP_temp;
	cfloat *stack[PYA_QS_STACK], **sptr = stack, *pm, *pi, *pj, *pt;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (NUMC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        if (NUMC_LT(*pr,*pm)) SWAP(*pr,*pm);
                        if (NUMC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        vp = *pm;
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (NUMC_LT(*pi,vp));
                                do --pj; while (NUMC_LT(vp,*pj));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && NUMC_LT(vp, *pt);) {
                                *pj-- = *pt--;
                        }
                        *pj = vp;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}

static int
CFLOAT_aquicksort(cfloat *v, intp* tosort, intp num, void *unused)
{
	cfloat vp;
	intp *pl, *pr, SWAP_temp;
	intp *stack[PYA_QS_STACK], **sptr=stack, *pm, *pi, *pj, *pt, vi;

	pl = tosort;
	pr = tosort + num - 1;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (NUMC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        if (NUMC_LT(v[*pr],v[*pm])) SWAP(*pr,*pm);
                        if (NUMC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        vp = v[*pm];
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (NUMC_LT(v[*pi],vp));
                                do --pj; while (NUMC_LT(vp,v[*pj]));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && NUMC_LT(vp, v[*pt]);)
				{
					*pj-- = *pt--;
				}
                        *pj = vi;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}


static int
CFLOAT_heapsort(cfloat *start, intp n, void *unused)
{
	
        cfloat tmp, *a;
        intp i,j,l;

        /* The array needs to be offset by one for heapsort indexing */
	a = start - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && NUMC_LT(a[j], a[j+1]))
                                j += 1;
                        if (NUMC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && NUMC_LT(a[j], a[j+1]))
                                j++;
                        if (NUMC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }
	return 0;
}

static int
CFLOAT_aheapsort(cfloat *v, intp *tosort, intp n, void *unused)
{
	intp *a, i,j,l, tmp;
	/* The arrays need to be offset by one for heapsort indexing */
	a = tosort - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && NUMC_LT(v[a[j]], v[a[j+1]]))
                                j += 1;
                        if (NUMC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && NUMC_LT(v[a[j]], v[a[j+1]]))
                                j++;
                        if (NUMC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

	return 0;
}

static void
CFLOAT_mergesort0(cfloat *pl, cfloat *pr, cfloat *pw)
{
	cfloat vp, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                CFLOAT_mergesort0(pl,pm-1,pw);
                CFLOAT_mergesort0(pm,pr,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (NUMC_LE(*pk,*pj)) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pk = pi - 1;\
			    pj > pl && NUMC_LT(vp, *pk); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vp;
                }
        }
}

static int
CFLOAT_mergesort(cfloat *start, intp num, void *unused)
{
	cfloat *pl, *pr, *pw;

	pl = start; pr = pl + num - 1;
	pw = (cfloat *) PyDataMem_NEW(((1+num/2))*sizeof(cfloat));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	CFLOAT_mergesort0(pl, pr, pw);
	PyDataMem_FREE(pw);
	return 0;
}

static void 
CFLOAT_amergesort0(intp *pl, intp *pr, cfloat *v, intp *pw)
{
        cfloat vp;
        intp vi, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                CFLOAT_amergesort0(pl,pm-1,v,pw);
                CFLOAT_amergesort0(pm,pr,v,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (NUMC_LE(v[*pk],v[*pj])) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pk = pi - 1;			\
			    pj > pl && NUMC_LT(vp, v[*pk]); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vi;
                }
        }
}

static int
CFLOAT_amergesort(cfloat *v, intp *tosort, intp num, void *unused)
{
	intp *pl, *pr, *pw;

	pl = tosort; pr = pl + num - 1;
	pw = PyDimMem_NEW((1+num/2));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	CFLOAT_amergesort0(pl, pr, v, pw);
	PyDimMem_FREE(pw);
	return 0;
}


static int
CDOUBLE_quicksort(cdouble *start, intp num, void *unused)
{
	cdouble *pl = start;
	cdouble *pr = start + num - 1;
	cdouble vp, SWAP_temp;
	cdouble *stack[PYA_QS_STACK], **sptr = stack, *pm, *pi, *pj, *pt;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (NUMC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        if (NUMC_LT(*pr,*pm)) SWAP(*pr,*pm);
                        if (NUMC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        vp = *pm;
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (NUMC_LT(*pi,vp));
                                do --pj; while (NUMC_LT(vp,*pj));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && NUMC_LT(vp, *pt);) {
                                *pj-- = *pt--;
                        }
                        *pj = vp;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}

static int
CDOUBLE_aquicksort(cdouble *v, intp* tosort, intp num, void *unused)
{
	cdouble vp;
	intp *pl, *pr, SWAP_temp;
	intp *stack[PYA_QS_STACK], **sptr=stack, *pm, *pi, *pj, *pt, vi;

	pl = tosort;
	pr = tosort + num - 1;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (NUMC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        if (NUMC_LT(v[*pr],v[*pm])) SWAP(*pr,*pm);
                        if (NUMC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        vp = v[*pm];
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (NUMC_LT(v[*pi],vp));
                                do --pj; while (NUMC_LT(vp,v[*pj]));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && NUMC_LT(vp, v[*pt]);)
				{
					*pj-- = *pt--;
				}
                        *pj = vi;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}


static int
CDOUBLE_heapsort(cdouble *start, intp n, void *unused)
{
	
        cdouble tmp, *a;
        intp i,j,l;

        /* The array needs to be offset by one for heapsort indexing */
	a = start - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && NUMC_LT(a[j], a[j+1]))
                                j += 1;
                        if (NUMC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && NUMC_LT(a[j], a[j+1]))
                                j++;
                        if (NUMC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }
	return 0;
}

static int
CDOUBLE_aheapsort(cdouble *v, intp *tosort, intp n, void *unused)
{
	intp *a, i,j,l, tmp;
	/* The arrays need to be offset by one for heapsort indexing */
	a = tosort - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && NUMC_LT(v[a[j]], v[a[j+1]]))
                                j += 1;
                        if (NUMC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && NUMC_LT(v[a[j]], v[a[j+1]]))
                                j++;
                        if (NUMC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

	return 0;
}

static void
CDOUBLE_mergesort0(cdouble *pl, cdouble *pr, cdouble *pw)
{
	cdouble vp, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                CDOUBLE_mergesort0(pl,pm-1,pw);
                CDOUBLE_mergesort0(pm,pr,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (NUMC_LE(*pk,*pj)) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pk = pi - 1;\
			    pj > pl && NUMC_LT(vp, *pk); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vp;
                }
        }
}

static int
CDOUBLE_mergesort(cdouble *start, intp num, void *unused)
{
	cdouble *pl, *pr, *pw;

	pl = start; pr = pl + num - 1;
	pw = (cdouble *) PyDataMem_NEW(((1+num/2))*sizeof(cdouble));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	CDOUBLE_mergesort0(pl, pr, pw);
	PyDataMem_FREE(pw);
	return 0;
}

static void 
CDOUBLE_amergesort0(intp *pl, intp *pr, cdouble *v, intp *pw)
{
        cdouble vp;
        intp vi, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                CDOUBLE_amergesort0(pl,pm-1,v,pw);
                CDOUBLE_amergesort0(pm,pr,v,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (NUMC_LE(v[*pk],v[*pj])) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pk = pi - 1;			\
			    pj > pl && NUMC_LT(vp, v[*pk]); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vi;
                }
        }
}

static int
CDOUBLE_amergesort(cdouble *v, intp *tosort, intp num, void *unused)
{
	intp *pl, *pr, *pw;

	pl = tosort; pr = pl + num - 1;
	pw = PyDimMem_NEW((1+num/2));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	CDOUBLE_amergesort0(pl, pr, v, pw);
	PyDimMem_FREE(pw);
	return 0;
}


static int
CLONGDOUBLE_quicksort(clongdouble *start, intp num, void *unused)
{
	clongdouble *pl = start;
	clongdouble *pr = start + num - 1;
	clongdouble vp, SWAP_temp;
	clongdouble *stack[PYA_QS_STACK], **sptr = stack, *pm, *pi, *pj, *pt;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (NUMC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        if (NUMC_LT(*pr,*pm)) SWAP(*pr,*pm);
                        if (NUMC_LT(*pm,*pl)) SWAP(*pm,*pl);
                        vp = *pm;
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (NUMC_LT(*pi,vp));
                                do --pj; while (NUMC_LT(vp,*pj));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && NUMC_LT(vp, *pt);) {
                                *pj-- = *pt--;
                        }
                        *pj = vp;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}

static int
CLONGDOUBLE_aquicksort(clongdouble *v, intp* tosort, intp num, void *unused)
{
	clongdouble vp;
	intp *pl, *pr, SWAP_temp;
	intp *stack[PYA_QS_STACK], **sptr=stack, *pm, *pi, *pj, *pt, vi;

	pl = tosort;
	pr = tosort + num - 1;

        for(;;) {
                while ((pr - pl) > SMALL_QUICKSORT) {
                        /* quicksort partition */
                        pm = pl + ((pr - pl) >> 1);
                        if (NUMC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        if (NUMC_LT(v[*pr],v[*pm])) SWAP(*pr,*pm);
                        if (NUMC_LT(v[*pm],v[*pl])) SWAP(*pm,*pl);
                        vp = v[*pm];
                        pi = pl;
                        pj = pr - 1;
                        SWAP(*pm,*pj);
                        for(;;) {
                                do ++pi; while (NUMC_LT(v[*pi],vp));
                                do --pj; while (NUMC_LT(vp,v[*pj]));
                                if (pi >= pj)  break;
                                SWAP(*pi,*pj);
                        }
                        SWAP(*pi,*(pr-1));
                        /* push largest partition on stack */
                        if (pi - pl < pr - pi) {
                                *sptr++ = pi + 1;
                                *sptr++ = pr;
                                pr = pi - 1;
                        }else{
                                *sptr++ = pl;
                                *sptr++ = pi - 1;
                                pl = pi + 1;
                        }
                }
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pt = pi - 1; \
			    pj > pl && NUMC_LT(vp, v[*pt]);)
				{
					*pj-- = *pt--;
				}
                        *pj = vi;
                }
                if (sptr == stack) break;
                pr = *(--sptr);
                pl = *(--sptr);
        }
	return 0;
}


static int
CLONGDOUBLE_heapsort(clongdouble *start, intp n, void *unused)
{
	
        clongdouble tmp, *a;
        intp i,j,l;

        /* The array needs to be offset by one for heapsort indexing */
	a = start - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && NUMC_LT(a[j], a[j+1]))
                                j += 1;
                        if (NUMC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && NUMC_LT(a[j], a[j+1]))
                                j++;
                        if (NUMC_LT(tmp, a[j])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }
	return 0;
}

static int
CLONGDOUBLE_aheapsort(clongdouble *v, intp *tosort, intp n, void *unused)
{
	intp *a, i,j,l, tmp;
	/* The arrays need to be offset by one for heapsort indexing */
	a = tosort - 1;

        for (l = n>>1; l > 0; --l) {
                tmp = a[l];
                for (i = l, j = l<<1; j <= n;) {
                        if (j < n && NUMC_LT(v[a[j]], v[a[j+1]]))
                                j += 1;
                        if (NUMC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

        for (; n > 1;) {
                tmp = a[n];
                a[n] = a[1];
                n -= 1;
                for (i = 1, j = 2; j <= n;) {
                        if (j < n && NUMC_LT(v[a[j]], v[a[j+1]]))
                                j++;
                        if (NUMC_LT(v[tmp], v[a[j]])) {
                                a[i] = a[j];
                                i = j;
                                j += j;
                        }else
                                break;
                }
                a[i] = tmp;
        }

	return 0;
}

static void
CLONGDOUBLE_mergesort0(clongdouble *pl, clongdouble *pr, clongdouble *pw)
{
	clongdouble vp, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                CLONGDOUBLE_mergesort0(pl,pm-1,pw);
                CLONGDOUBLE_mergesort0(pm,pr,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (NUMC_LE(*pk,*pj)) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vp = *pi;
                        for(pj = pi, pk = pi - 1;\
			    pj > pl && NUMC_LT(vp, *pk); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vp;
                }
        }
}

static int
CLONGDOUBLE_mergesort(clongdouble *start, intp num, void *unused)
{
	clongdouble *pl, *pr, *pw;

	pl = start; pr = pl + num - 1;
	pw = (clongdouble *) PyDataMem_NEW(((1+num/2))*sizeof(clongdouble));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	CLONGDOUBLE_mergesort0(pl, pr, pw);
	PyDataMem_FREE(pw);
	return 0;
}

static void 
CLONGDOUBLE_amergesort0(intp *pl, intp *pr, clongdouble *v, intp *pw)
{
        clongdouble vp;
        intp vi, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                CLONGDOUBLE_amergesort0(pl,pm-1,v,pw);
                CLONGDOUBLE_amergesort0(pm,pr,v,pw);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (NUMC_LE(v[*pk],v[*pj])) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pk = pi - 1;			\
			    pj > pl && NUMC_LT(vp, v[*pk]); --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vi;
                }
        }
}

static int
CLONGDOUBLE_amergesort(clongdouble *v, intp *tosort, intp num, void *unused)
{
	intp *pl, *pr, *pw;

	pl = tosort; pr = pl + num - 1;
	pw = PyDimMem_NEW((1+num/2));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	CLONGDOUBLE_amergesort0(pl, pr, v, pw);
	PyDimMem_FREE(pw);
	return 0;
}



static int
unincmp(Py_UNICODE *s1, Py_UNICODE *s2, register int len)
{
	register Py_UNICODE c1, c2;
	while(len-- > 0) {
		c1 = *s1++;
		c2 = *s2++;
		if (c1 != c2)
			return (c1 < c2) ? -1 : 1;
	}
	return 0;
}

static void 
STRING_amergesort0(intp *pl, intp *pr, char **v, intp *pw, int elsize)
{
	char * vp;
        intp vi, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                STRING_amergesort0(pl,pm-1,v,pw,elsize);
                STRING_amergesort0(pm,pr,v,pw,elsize);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (strncmp(v[*pk],v[*pj],elsize)<=0) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pk = pi - 1; 		\
			    pj > pl && (strncmp(vp, v[*pk],elsize)<=0); \
			    --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vi;
                }
        }
}

static int
STRING_amergesort(char **v, intp *tosort, intp num, PyArrayObject *arr)
{
	intp *pl, *pr, *pw;
	int elsize;

	elsize = arr->descr->elsize;

	pl = tosort; pr = pl + num - 1;
	pw = PyDimMem_NEW((1+num/2));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	STRING_amergesort0(pl, pr, v, pw, elsize);
	PyDimMem_FREE(pw);
	return 0;
}


static void 
UNICODE_amergesort0(intp *pl, intp *pr, Py_UNICODE **v, intp *pw, int elsize)
{
	Py_UNICODE * vp;
        intp vi, *pi, *pj, *pk, *pm;

        if (pr - pl > SMALL_MERGESORT) {
                /* merge sort */
                pm = pl + ((pr - pl + 1)>>1);
                UNICODE_amergesort0(pl,pm-1,v,pw,elsize);
                UNICODE_amergesort0(pm,pr,v,pw,elsize);
                for(pi = pw, pj = pl; pj < pm; ++pi, ++pj) {
                        *pi = *pj;
                }
                for(pk = pw, pm = pl; pk < pi && pj <= pr; ++pm) {
                        if (unincmp(v[*pk],v[*pj],elsize)<=0) {
                                *pm = *pk;
                                ++pk;
                        }else{
                                *pm = *pj;
                                ++pj;
                        }
                }
                for(; pk < pi; ++pm, ++pk) {
                        *pm = *pk;
                }
        }else{
                /* insertion sort */
                for(pi = pl + 1; pi <= pr; ++pi) {
                        vi = *pi;
                        vp = v[vi];
                        for(pj = pi, pk = pi - 1; 		\
			    pj > pl && (unincmp(vp, v[*pk],elsize)<=0); \
			    --pj, --pk) {
                                *pj = *pk;
                        }
                        *pj = vi;
                }
        }
}

static int
UNICODE_amergesort(Py_UNICODE **v, intp *tosort, intp num, PyArrayObject *arr)
{
	intp *pl, *pr, *pw;
	int elsize;

	elsize = arr->descr->elsize;

	pl = tosort; pr = pl + num - 1;
	pw = PyDimMem_NEW((1+num/2));
	
	if (!pw) {
		PyErr_NoMemory();
		return -1;
	}

	UNICODE_amergesort0(pl, pr, v, pw, elsize);
	PyDimMem_FREE(pw);
	return 0;
}


	
static void 
add_sortfuncs(void)
{
	PyArray_Descr *descr;

	descr = PyArray_DescrFromType(PyArray_BOOL);
	descr->f->sort[PyArray_QUICKSORT] = \
		(PyArray_SortFunc *)BOOL_quicksort;
	descr->f->argsort[PyArray_QUICKSORT] = \
		(PyArray_ArgSortFunc *)BOOL_aquicksort;
	descr->f->sort[PyArray_HEAPSORT] = \
		(PyArray_SortFunc *)BOOL_heapsort;
	descr->f->argsort[PyArray_HEAPSORT] = \
		(PyArray_ArgSortFunc *)BOOL_aheapsort;
	descr->f->sort[PyArray_MERGESORT] = \
		(PyArray_SortFunc *)BOOL_mergesort;
	descr->f->argsort[PyArray_MERGESORT] = \
		(PyArray_ArgSortFunc *)BOOL_amergesort;


	descr = PyArray_DescrFromType(PyArray_BYTE);
	descr->f->sort[PyArray_QUICKSORT] = \
		(PyArray_SortFunc *)BYTE_quicksort;
	descr->f->argsort[PyArray_QUICKSORT] = \
		(PyArray_ArgSortFunc *)BYTE_aquicksort;
	descr->f->sort[PyArray_HEAPSORT] = \
		(PyArray_SortFunc *)BYTE_heapsort;
	descr->f->argsort[PyArray_HEAPSORT] = \
		(PyArray_ArgSortFunc *)BYTE_aheapsort;
	descr->f->sort[PyArray_MERGESORT] = \
		(PyArray_SortFunc *)BYTE_mergesort;
	descr->f->argsort[PyArray_MERGESORT] = \
		(PyArray_ArgSortFunc *)BYTE_amergesort;


	descr = PyArray_DescrFromType(PyArray_UBYTE);
	descr->f->sort[PyArray_QUICKSORT] = \
		(PyArray_SortFunc *)UBYTE_quicksort;
	descr->f->argsort[PyArray_QUICKSORT] = \
		(PyArray_ArgSortFunc *)UBYTE_aquicksort;
	descr->f->sort[PyArray_HEAPSORT] = \
		(PyArray_SortFunc *)UBYTE_heapsort;
	descr->f->argsort[PyArray_HEAPSORT] = \
		(PyArray_ArgSortFunc *)UBYTE_aheapsort;
	descr->f->sort[PyArray_MERGESORT] = \
		(PyArray_SortFunc *)UBYTE_mergesort;
	descr->f->argsort[PyArray_MERGESORT] = \
		(PyArray_ArgSortFunc *)UBYTE_amergesort;


	descr = PyArray_DescrFromType(PyArray_SHORT);
	descr->f->sort[PyArray_QUICKSORT] = \
		(PyArray_SortFunc *)SHORT_quicksort;
	descr->f->argsort[PyArray_QUICKSORT] = \
		(PyArray_ArgSortFunc *)SHORT_aquicksort;
	descr->f->sort[PyArray_HEAPSORT] = \
		(PyArray_SortFunc *)SHORT_heapsort;
	descr->f->argsort[PyArray_HEAPSORT] = \
		(PyArray_ArgSortFunc *)SHORT_aheapsort;
	descr->f->sort[PyArray_MERGESORT] = \
		(PyArray_SortFunc *)SHORT_mergesort;
	descr->f->argsort[PyArray_MERGESORT] = \
		(PyArray_ArgSortFunc *)SHORT_amergesort;


	descr = PyArray_DescrFromType(PyArray_USHORT);
	descr->f->sort[PyArray_QUICKSORT] = \
		(PyArray_SortFunc *)USHORT_quicksort;
	descr->f->argsort[PyArray_QUICKSORT] = \
		(PyArray_ArgSortFunc *)USHORT_aquicksort;
	descr->f->sort[PyArray_HEAPSORT] = \
		(PyArray_SortFunc *)USHORT_heapsort;
	descr->f->argsort[PyArray_HEAPSORT] = \
		(PyArray_ArgSortFunc *)USHORT_aheapsort;
	descr->f->sort[PyArray_MERGESORT] = \
		(PyArray_SortFunc *)USHORT_mergesort;
	descr->f->argsort[PyArray_MERGESORT] = \
		(PyArray_ArgSortFunc *)USHORT_amergesort;


	descr = PyArray_DescrFromType(PyArray_INT);
	descr->f->sort[PyArray_QUICKSORT] = \
		(PyArray_SortFunc *)INT_quicksort;
	descr->f->argsort[PyArray_QUICKSORT] = \
		(PyArray_ArgSortFunc *)INT_aquicksort;
	descr->f->sort[PyArray_HEAPSORT] = \
		(PyArray_SortFunc *)INT_heapsort;
	descr->f->argsort[PyArray_HEAPSORT] = \
		(PyArray_ArgSortFunc *)INT_aheapsort;
	descr->f->sort[PyArray_MERGESORT] = \
		(PyArray_SortFunc *)INT_mergesort;
	descr->f->argsort[PyArray_MERGESORT] = \
		(PyArray_ArgSortFunc *)INT_amergesort;


	descr = PyArray_DescrFromType(PyArray_UINT);
	descr->f->sort[PyArray_QUICKSORT] = \
		(PyArray_SortFunc *)UINT_quicksort;
	descr->f->argsort[PyArray_QUICKSORT] = \
		(PyArray_ArgSortFunc *)UINT_aquicksort;
	descr->f->sort[PyArray_HEAPSORT] = \
		(PyArray_SortFunc *)UINT_heapsort;
	descr->f->argsort[PyArray_HEAPSORT] = \
		(PyArray_ArgSortFunc *)UINT_aheapsort;
	descr->f->sort[PyArray_MERGESORT] = \
		(PyArray_SortFunc *)UINT_mergesort;
	descr->f->argsort[PyArray_MERGESORT] = \
		(PyArray_ArgSortFunc *)UINT_amergesort;


	descr = PyArray_DescrFromType(PyArray_LONG);
	descr->f->sort[PyArray_QUICKSORT] = \
		(PyArray_SortFunc *)LONG_quicksort;
	descr->f->argsort[PyArray_QUICKSORT] = \
		(PyArray_ArgSortFunc *)LONG_aquicksort;
	descr->f->sort[PyArray_HEAPSORT] = \
		(PyArray_SortFunc *)LONG_heapsort;
	descr->f->argsort[PyArray_HEAPSORT] = \
		(PyArray_ArgSortFunc *)LONG_aheapsort;
	descr->f->sort[PyArray_MERGESORT] = \
		(PyArray_SortFunc *)LONG_mergesort;
	descr->f->argsort[PyArray_MERGESORT] = \
		(PyArray_ArgSortFunc *)LONG_amergesort;


	descr = PyArray_DescrFromType(PyArray_ULONG);
	descr->f->sort[PyArray_QUICKSORT] = \
		(PyArray_SortFunc *)ULONG_quicksort;
	descr->f->argsort[PyArray_QUICKSORT] = \
		(PyArray_ArgSortFunc *)ULONG_aquicksort;
	descr->f->sort[PyArray_HEAPSORT] = \
		(PyArray_SortFunc *)ULONG_heapsort;
	descr->f->argsort[PyArray_HEAPSORT] = \
		(PyArray_ArgSortFunc *)ULONG_aheapsort;
	descr->f->sort[PyArray_MERGESORT] = \
		(PyArray_SortFunc *)ULONG_mergesort;
	descr->f->argsort[PyArray_MERGESORT] = \
		(PyArray_ArgSortFunc *)ULONG_amergesort;


	descr = PyArray_DescrFromType(PyArray_LONGLONG);
	descr->f->sort[PyArray_QUICKSORT] = \
		(PyArray_SortFunc *)LONGLONG_quicksort;
	descr->f->argsort[PyArray_QUICKSORT] = \
		(PyArray_ArgSortFunc *)LONGLONG_aquicksort;
	descr->f->sort[PyArray_HEAPSORT] = \
		(PyArray_SortFunc *)LONGLONG_heapsort;
	descr->f->argsort[PyArray_HEAPSORT] = \
		(PyArray_ArgSortFunc *)LONGLONG_aheapsort;
	descr->f->sort[PyArray_MERGESORT] = \
		(PyArray_SortFunc *)LONGLONG_mergesort;
	descr->f->argsort[PyArray_MERGESORT] = \
		(PyArray_ArgSortFunc *)LONGLONG_amergesort;


	descr = PyArray_DescrFromType(PyArray_ULONGLONG);
	descr->f->sort[PyArray_QUICKSORT] = \
		(PyArray_SortFunc *)ULONGLONG_quicksort;
	descr->f->argsort[PyArray_QUICKSORT] = \
		(PyArray_ArgSortFunc *)ULONGLONG_aquicksort;
	descr->f->sort[PyArray_HEAPSORT] = \
		(PyArray_SortFunc *)ULONGLONG_heapsort;
	descr->f->argsort[PyArray_HEAPSORT] = \
		(PyArray_ArgSortFunc *)ULONGLONG_aheapsort;
	descr->f->sort[PyArray_MERGESORT] = \
		(PyArray_SortFunc *)ULONGLONG_mergesort;
	descr->f->argsort[PyArray_MERGESORT] = \
		(PyArray_ArgSortFunc *)ULONGLONG_amergesort;


	descr = PyArray_DescrFromType(PyArray_FLOAT);
	descr->f->sort[PyArray_QUICKSORT] = \
		(PyArray_SortFunc *)FLOAT_quicksort;
	descr->f->argsort[PyArray_QUICKSORT] = \
		(PyArray_ArgSortFunc *)FLOAT_aquicksort;
	descr->f->sort[PyArray_HEAPSORT] = \
		(PyArray_SortFunc *)FLOAT_heapsort;
	descr->f->argsort[PyArray_HEAPSORT] = \
		(PyArray_ArgSortFunc *)FLOAT_aheapsort;
	descr->f->sort[PyArray_MERGESORT] = \
		(PyArray_SortFunc *)FLOAT_mergesort;
	descr->f->argsort[PyArray_MERGESORT] = \
		(PyArray_ArgSortFunc *)FLOAT_amergesort;


	descr = PyArray_DescrFromType(PyArray_DOUBLE);
	descr->f->sort[PyArray_QUICKSORT] = \
		(PyArray_SortFunc *)DOUBLE_quicksort;
	descr->f->argsort[PyArray_QUICKSORT] = \
		(PyArray_ArgSortFunc *)DOUBLE_aquicksort;
	descr->f->sort[PyArray_HEAPSORT] = \
		(PyArray_SortFunc *)DOUBLE_heapsort;
	descr->f->argsort[PyArray_HEAPSORT] = \
		(PyArray_ArgSortFunc *)DOUBLE_aheapsort;
	descr->f->sort[PyArray_MERGESORT] = \
		(PyArray_SortFunc *)DOUBLE_mergesort;
	descr->f->argsort[PyArray_MERGESORT] = \
		(PyArray_ArgSortFunc *)DOUBLE_amergesort;


	descr = PyArray_DescrFromType(PyArray_LONGDOUBLE);
	descr->f->sort[PyArray_QUICKSORT] = \
		(PyArray_SortFunc *)LONGDOUBLE_quicksort;
	descr->f->argsort[PyArray_QUICKSORT] = \
		(PyArray_ArgSortFunc *)LONGDOUBLE_aquicksort;
	descr->f->sort[PyArray_HEAPSORT] = \
		(PyArray_SortFunc *)LONGDOUBLE_heapsort;
	descr->f->argsort[PyArray_HEAPSORT] = \
		(PyArray_ArgSortFunc *)LONGDOUBLE_aheapsort;
	descr->f->sort[PyArray_MERGESORT] = \
		(PyArray_SortFunc *)LONGDOUBLE_mergesort;
	descr->f->argsort[PyArray_MERGESORT] = \
		(PyArray_ArgSortFunc *)LONGDOUBLE_amergesort;


	descr = PyArray_DescrFromType(PyArray_CFLOAT);
	descr->f->sort[PyArray_QUICKSORT] = \
		(PyArray_SortFunc *)CFLOAT_quicksort;
	descr->f->argsort[PyArray_QUICKSORT] = \
		(PyArray_ArgSortFunc *)CFLOAT_aquicksort;
	descr->f->sort[PyArray_HEAPSORT] = \
		(PyArray_SortFunc *)CFLOAT_heapsort;
	descr->f->argsort[PyArray_HEAPSORT] = \
		(PyArray_ArgSortFunc *)CFLOAT_aheapsort;
	descr->f->sort[PyArray_MERGESORT] = \
		(PyArray_SortFunc *)CFLOAT_mergesort;
	descr->f->argsort[PyArray_MERGESORT] = \
		(PyArray_ArgSortFunc *)CFLOAT_amergesort;


	descr = PyArray_DescrFromType(PyArray_CDOUBLE);
	descr->f->sort[PyArray_QUICKSORT] = \
		(PyArray_SortFunc *)CDOUBLE_quicksort;
	descr->f->argsort[PyArray_QUICKSORT] = \
		(PyArray_ArgSortFunc *)CDOUBLE_aquicksort;
	descr->f->sort[PyArray_HEAPSORT] = \
		(PyArray_SortFunc *)CDOUBLE_heapsort;
	descr->f->argsort[PyArray_HEAPSORT] = \
		(PyArray_ArgSortFunc *)CDOUBLE_aheapsort;
	descr->f->sort[PyArray_MERGESORT] = \
		(PyArray_SortFunc *)CDOUBLE_mergesort;
	descr->f->argsort[PyArray_MERGESORT] = \
		(PyArray_ArgSortFunc *)CDOUBLE_amergesort;


	descr = PyArray_DescrFromType(PyArray_CLONGDOUBLE);
	descr->f->sort[PyArray_QUICKSORT] = \
		(PyArray_SortFunc *)CLONGDOUBLE_quicksort;
	descr->f->argsort[PyArray_QUICKSORT] = \
		(PyArray_ArgSortFunc *)CLONGDOUBLE_aquicksort;
	descr->f->sort[PyArray_HEAPSORT] = \
		(PyArray_SortFunc *)CLONGDOUBLE_heapsort;
	descr->f->argsort[PyArray_HEAPSORT] = \
		(PyArray_ArgSortFunc *)CLONGDOUBLE_aheapsort;
	descr->f->sort[PyArray_MERGESORT] = \
		(PyArray_SortFunc *)CLONGDOUBLE_mergesort;
	descr->f->argsort[PyArray_MERGESORT] = \
		(PyArray_ArgSortFunc *)CLONGDOUBLE_amergesort;



	descr = PyArray_DescrFromType(PyArray_STRING);
	descr->f->argsort[PyArray_MERGESORT] = \
		(PyArray_ArgSortFunc *)STRING_amergesort;
	descr = PyArray_DescrFromType(PyArray_UNICODE);
	descr->f->argsort[PyArray_MERGESORT] = \
		(PyArray_ArgSortFunc *)UNICODE_amergesort;
}

static struct PyMethodDef methods[] = {
	{NULL, NULL, 0}
};

PyMODINIT_FUNC
init_sort(void) {
	PyObject *m;

	m = Py_InitModule("_sort", methods);
	
	if (import_array() < 0) return;
	add_sortfuncs();
}
